//------------------------------------------------------------------------------
//   hash.cpp           Hash-Algorithmen
//   zuletzt gendert   am:   10.12.01/Kp
//------------------------------------------------------------------------------
//	void 	InitHTable      Hash-Tabelle initialisieren
//	int 	Hash            Hash-Funktion
//	void 	InsertItem      neues Item in die Hash-Tabelle eintragen
//	void 	DeleteItem      Item aus der Hashtabelle lschen
//	void 	HSearch         Item in der Hashtabelle suchen
//	void 	ReHash          Hashtabelle dynamisch vergrern, verkleinern
//------------------------------------------------------------------------------

#include <iostream.h>
#include <stdlib.h>
#include <limits.h>
#include <conio>
#include <math.h>
#include <iomanip.h>
#include "hash.h"

//------------------------------------------------------------------------------
//  HTable::HashM  :  Berechnet Hash-Index nach der Multtiplikationsmethode
//------------------------------------------------------------------------------
int HTable::Hash (int key)
{   double f;
	 f = 0.5 * (sqrt(5) - 1);    			     // Koeffizient des goldenen Schnitts
	 return (int) (hsize * (f*key - (int)(f*key))); 

    // andere Hashfunktionen:
    // ----------------------
    // Hash-Index nach dem Divisonsrest-Verfahren bestimmen
    // Tabellenlaengen sind 2^3, 2^4, 2^5, Die naechst kleiner Primzahl: 2^n - 1
    // return  (key % (hsize-1));

    // Hash-Index mit Multiplikationsmethode und Schlsselfaltung
    // double f;
    // key = key / 10000 + key % 10000;          // Faltung mit zwei Bloecken zu 4 Ziffern
    // f = 0.5 * (sqrt(5) - 1);    			     // Koeffizient des goldenen Schnitts
    // key = hsize * (f*key - (unsigned int)(f*key)); 	
    // return (key);

    // Basistransformation
    // char key8[20];
    // itoa (key, key8, 8);
    // cout << atol (key8) << endl;
    // return (((unsigned int)atol(key8) % (hsize-1)));
}

//------------------------------------------------------------------------------
//  Konstruktor fr die Klasse HTable
//------------------------------------------------------------------------------
HTable::HTable()
{
	tsize = 8;    								 // Standardgre der Hash-Tabelle
	step  = 1;                    		         // Standardgre fr Ausweichschritte
	table = new Item[tsize];      		         // Platz fr die Tabelle reservieren
   if ( table == NULL)
	{	cout << "***** Hash-Tabelle kann nicht angelegt werden !\n";
		return;
	}
	// Header initialisieren
	hsize    = tsize;							 // Tabellengre in Anzahl von Items
	ItemNum	= 0;								 // Anzahl belegter Positionen
												 // Alle key-Werte auf 0 setzen
	for (int i=0; i<tsize; i++) table[i].key = 0;
}

//------------------------------------------------------------------------------
//  PrintHtable:  Kontrollausgabe der Hashtabelle
//------------------------------------------------------------------------------
void 	HTable::PrintHtable ()
{
    if (table != NULL)
 	  for (int i=0; i<hsize; i++)
		cout << setw(5) << i << ":   " << setw(10) << table[i].key 
                                       << "   "    << table[i].data << endl;
    else
   	  cout << "Tabelle ist leer.\n";
}

//------------------------------------------------------------------------------
//  HSearch    :  Sucht ein Item in der Hash-Tabelle
//  Parameter  :  int    key       Schlssel des zu suchenden Item
//                int    &index    Index des gefundenen Item
//                int    &found    TRUE falls Item gefunden wurde
//------------------------------------------------------------------------------
void HTable::HSearch (int key, int &index, int &found)
{	int Hindex;

	found = false;                               // vorsorglich: Fehlanzeige einstellen
	if (ItemNum > 0)                             // falls Tabelle berhaupt Items enthlt
	{
	  Hindex = Hash (key);                       // Hash-Index berechnen
	  index = Hindex;                            // in Hindex bleibt die Ausgangsposition
	  if ( table[index].key != key )            // mit index durch die Tabelle laufen
	    do                                       // bis Item gefunden oder sicher nicht da ist
		  index = (index + step) % hsize;
		while ( table[index].key != key  &&  index != Hindex );
		if ( table[index].key == key ) found = true;
	}
	else
	  cout << "Tabelle enthaelt keine Items !\n";
}

//------------------------------------------------------------------------------
//  DeleteItem : 	Lscht ein Item mit dem Schlssel "key" aus der Hash-Tabelle
//------------------------------------------------------------------------------
void 	HTable::DeleteItem (int key)
{
	int index;
	int found;

	if (ItemNum == 0) cout << "Tabelle enthaelt keine Items !\n";
	else 										 // jetzt ist die Tabelle nicht leer
	{
		HSearch (key, index, found);             // Item suchen
		if (found == false)                      // kein Item gefunden: nur Fehlanzeige
			cout << "Item wurde nicht gefunden !\n";
		else                                     // jetzt steht das Item auf Position index
		{
			table[index].key = 0;                // Tabellenposition freigeben
			table[index].data[0] = 0;            // Tabellenposition freigeben
			ItemNum --;                          // ItemNum aktualisieren
			cout << "Item %d wurde geloescht !\n", key;
			if ( ((float)ItemNum / (float)hsize <= 0.5) && (hsize > tsize) )
				ReHash (-1);                     // falls erforderlich Tabelle verkleinern
			if (ItemNum == 0)                    // falls erforderlich Tabelle ganz freigeben
			{
				delete[] table;
				table = NULL;
			}
		}
	}
}

//------------------------------------------------------------------------------
//  InsertItem :  fgt ein Item in die Hash-Tabelle ein
//                Mehfarche Schluessel werden abgelehnt
//  Parameter  :  Item   item  Einzutragendes Item
//------------------------------------------------------------------------------
void 	HTable::InsertItem (Item item)
{
	int Hindex;                                  // Tabellenindex
	int pos;                                     // hlt eine Tabellenposition fest

	if (ItemNum >= 4*tsize)                      // Tabelle nicht mehr erweiterbar
		cout << "***** Hashtabelle ist voll. Kein weiteres Item eingefuegt !\n";
	else
	{
		// Tabelle so weit gefllt, da erweitert werden mu
		if ( ItemNum/hsize >= 0.9 && hsize <= 2*tsize )
			ReHash (1);

		Hindex = Hash(item.key);                 // Hash-Index bestimmen
		pos    = hsize+1;						 // Position auerhalb merken
		if (table[Hindex].key != 0)
		{
			pos = Hindex;                        // pos = Wunschziel
			do
            {                                   
              if (table[Hindex].key == item.key)
         	  {
         		cout << "Schluessel \"" << item.data << "\" ist bereits vorhanden !\n";
           	    return;
         	  }
              // Ausweichposition suchen
              Hindex = (Hindex + step) % hsize;
            } while (table[Hindex].key != 0 && Hindex!=pos);
		}
		if (Hindex == pos)                       // falls keine Ausweichposition da war
			cout << "***** Keine freie Position gefunden\n";
		else                                     // falls der Hash-Index O.K. war oder
		{                                        // Ausweichposition gefunden wurde
			ItemNum ++;                          // Anzahl der Items aktualisieren
			table[Hindex].key = item.key;        // Item eintragen
         strcpy ((table[Hindex]).data, item.data);
		}
	}
}


//------------------------------------------------------------------------------
//  ReHash     : 	Pat die Hash-Tabelle dynamisch an
//  Parameter  :  int   flag   +1 : Tabelle erweitern
//                             -1 : Tabelle verkrzen
//------------------------------------------------------------------------------
void 	HTable::ReHash (int flag)
{
	int Hsize;						             // Hilfsvariable fr die Tabellengre
	int i;							             // Index-Variable
	int HItemNum;					             // Hilfsvariable fr die Anzahl von Items
	Item * Htab;                                 // Hilfszeiger zur alten Tabelle

	Htab 		= table;				         // Zeiger zur Originaltabelle
	Hsize 	    = hsize;				         // alte Tabellengre
	HItemNum	= ItemNum;			             // alte Anzahl von Items in der Tabelle

	// Neue Tabelle hat die Gre 2*tsize
	if ( (flag == 1 && hsize == tsize) || (flag == -1 && hsize == 4*tsize) )
	{
	  // Platz fr die Tabelle reservieren
      table = new Item[2*tsize];
		if ( table == NULL)
		{	cout << "***** Hash-Tabelle kann nicht angelegt werden !\n";
			return;
		}
		// Header initialisieren
		hsize   = 2*tsize;                       // Tabellengre in Anzahl von Items
		step 	= 5*step;                        // Neue Schrittweite
	}
	// Neue Tabelle hat die Gre 3*tsize
	else if (flag == 1 && hsize == 2*tsize)
	{
		// Platz fr die Tabelle reservieren
      table = new Item[4*tsize];
		if ( table == NULL)
		{	cout << "***** Hash-Tabelle kann nicht angelegt werden !\n";
			return;
		}
		// Header initialisieren
		hsize   = 4*tsize;                       // Tabellengre in Anzahl von Items
		step 	= 7*step;                        // Neue Schrittweite
	}
	// Neue Tabelle hat die Gre tsize
	else if (flag == -1 && hsize == 2*tsize)
	{
	  // Platz fr die Tabelle reservieren
      table = new Item[tsize];
	  if ( table == NULL)
	  {	cout << "***** Hash-Tabelle kann nicht angelegt werden !\n";
	  	return;
	  }
	  // Header initialisieren
	  hsize   = tsize;    	                     // Tabellengre in Anzahl von Items
	  step 			= 3*step;     	             // Neue Schrittweite
	}
	// Falscher Aufruf: darf eigentlich nicht vorkommen
	else cout << "Fehlerhafter ReHash-Aufruf !\n";

	// In alle Items der neuen Tabelle den Key 0 eintragen
	for (i=0; i<hsize; i++) table[i].key = 0;

	// Daten aus der alten Tabelle in die neue bernehmen
	ItemNum = 0;
	for (i=0; i<=Hsize-1; i++)
	  if (Htab[i].key != 0)
	  {
	    InsertItem (Htab[i]);
	    HItemNum --;
	  }
	if (HItemNum != 0)
	  cout << "***** Es wurden nicht alle Items in die neue Tabelle uebernommen !\n";
	  delete[] Htab;						     // alte Tabelle freigeben;
}

//------------------------------------------------------------------------------
//  numericKey : 	berechnet aus einem String-Schlssel einen Integer-Schlssel
//					erlaubte String-Zeichen sind A,B, ... ,Z,a,b, ... ,z
//------------------------------------------------------------------------------
unsigned int  HTable::numericKey (char *stringKey)
{
    unsigned int key = 0;
    int c;

    for (int i=strlen(stringKey) -1 ; i>=0; i--)
    {
      if (    ((stringKey[i] >= '0') && (stringKey[i] <= '9'))
   	       || ((stringKey[i] >= 'A') && (stringKey[i] <= 'Z'))
   	       || ((stringKey[i] >= 'a') && (stringKey[i] <= 'z'))  )
      {
        if      (stringKey[i] <= '9')   c = stringKey[i] - '0' + 1;
        else if (stringKey[i] <= 'Z')   c = stringKey[i] - '0' -  7 + 1;
        else                            c = stringKey[i] - '0' - 13 + 1;
        key = key * (long)63 + (long)c;
      }
      else return (0);
    }
    //cout << key;
    return(key);
}


