//==============================================================================
//   SuchAlgorithmen
//   zuletzt gendert am:   29.01.01/Kp
//==============================================================================
//
//  Implementierte Funktionen:
//  Select   		Sucht das i-kleinste Item einer Liste rekursiv
//  SelectNR   	    Sucht das i-kleinste Item einer Liste nichtrekursiv
//  SearchSeq 		Sequentielles Suchen in einem Array
//  SearchBin 		Binres Suchen in einem Array
//  SearchTxt   	Direkte Mustersuche in einem Text
//  SearchKMP_O	    Mustersuche nach den Knuth-Morris-Pratt-Algorithmus (Ottmann)
//  SearchKMP_S	    Mustersuche nach den Knuth-Morris-Pratt-Algorithmus (Sedgewick)
//  LvDist          Levenstein-Distatnz zweier Zeichenfolgen
//  SearchLv        Unscharfe Textmustersuche nach Levenstein
//
//==============================================================================

#include <string.h>
#include "Search.h"
#include <iostream.h>

//------------------------------------------------------------------------------
//  Select:  Sucht das     i-kleinste Item einer Liste rekursiv
//  Parameter:    A        Zahlenfeld, das zu zerlegen ist
//                left     Index der linken Grenze
//                right    Index der rechten Grenze
//                k        Es wird das k-kleinste gesucht
//                found    Wert des k-kleinsten Elements
//------------------------------------------------------------------------------
void Select (int A[], int left, int right, int i, int &found)
{
    int	m;									// Position des teilenden Elements

    if (right>left) 						// Falls Listenlnge >=2
    {
      m = Partition (A,left,right); 	    // Liste aufteilen
      if (i<=m-left)                	    // falls Item noch nicht gefunden
        Select (A,left,m-1,i,found);        // suche im linken Teil weiter
      else
        Select (A,m,right,i-(m-left),found);// suche im rechten Teil weiter
	}
   else
   	found =A[left];						    // Item gefunden, Wert abliefern
	return;
}

//------------------------------------------------------------------------------
//  SelectNR:     Sucht das i-kleinste Item einer Liste nichtrekursiv
//  Parameter:    A        Zahlenfeld, das zu zerlegen ist
//                left     Index der linken Grenze
//                right    Index der rechten Grenze
//                k        Es wird das k-kleinste gesucht
//                found    Wert des k-kleinsten Elements
//------------------------------------------------------------------------------
void SelectNR (int *A, int left, int right, int i, int &found)
{
	int	j, k;								// Laufvariablen
	int	item;			 					// Hilfsfeld zum Vertauschen
	int	pivot;								// Pivot-Element

	while (left<right)      				// Falls Listenlnge >=2
	{
		pivot = A[i];             			// k-tes Item wird Pivot
		k = left; j = right;        		// Laufzeiger initialisieren
		do
		{
          while ((A[k] < pivot)) k++;  	    // von links  her ein "grosses" El. suchen
          while ((A[j] > pivot)) j--;  	    // von rechts her ein "kleines" El. suchen
          if (k <= j)
          {
            item = A[k];                    // grosses und kleines Element vertauschen
            A[k] = A[j];
            A[j] = item;
            k++; j--;
          }
		}  while (k<=j);

		// Falls das linke Teilfeld zu kurz ist, suche rechts weiter
		if (j<i)  left  = k;            	// suche im rechten Teil weiter
		if (i<k)  right = j;				// suche im rechten Teil weiter
	}
	found = A[i];							// Item gefunden, Wert abliefern
	return;
}

//------------------------------------------------------------------------------
//  Partition:    Zerlegen der Liste mit einem Pivot-Element
//  Parameter:    A        Zahlenfeld, das zu zerlegen ist
//                left     Index der linken Grenze
//                right    Index der rechten Grenze
//------------------------------------------------------------------------------
int  Partition  (int *A, int left, int right)
{
	int	i, j;								// Laufvariablen
	int	item;			 					// Hilfsfeld zum Vertauschene
	int	pivot;							    // Pivot-Element

	pivot = A[right];             	        // Pivot-Element festhalten
	i = left-1, j = right;        	        // Laufzeiger initialisieren
	for (;;)
	{
		while ((A[++i] < pivot));    	    // von links  her ein "grosses" El. suchen
		while ((A[--j] > pivot));    	    // von rechts her ein "kleines" El. suchen
		if (i >= j) break;         	        // Abbruch, wenn sich die Zeiger ueberholen
		item = A[i];               	        // grosses und kleines Element vertauschen
		A[i] = A[j];
		A[j] = item;
	}
	A[right] = A[i] ;         			    // Pivot-Element auf endgltigen Platz
	A[i] = pivot;
	return(i);
}

//------------------------------------------------------------------------------
//  SearchSeq:	  Sequentielles Suchen in einem Array
//  Parameter:    A     Zahlenfeld, in dem gesucht wird
//                n     Anzahl der gltigen Feldelemente
//                item  gesuchtes Element
//------------------------------------------------------------------------------
int  SearchSeq  (int *A, int n, int item)
{
	int i;									// Laufvariable
	A[0] = item;							// Stopper-Element besetzen
	i    = n + 1;              			    // An das rechte Ende zeigen
	while (item != A[--i]);    			    // Von rechts suchen
	return (i);								// Index des Item oder 0 zurckliefern
}

//------------------------------------------------------------------------------
//  SearchBin:	  Binres Suchen in einem Array
//  Parameter:    A     Zahlenfeld, in dem gesucht wird
//                n     Anzahl der gltigen Feldelemente
//                item  gesuchtes Element
//------------------------------------------------------------------------------
int  SearchBin  (int *A, int n, int item)
{
	int i, left, right;

   left=1; right=n;
	while (right >= left)
	{
		i = (left + right) / 2;
		if (item < A[i]) 	right = i-1;    // links weitersuchen
		else              left  = i+1;      // rechts weitersuchen
		if (item == A[i]) return(i);        // Treffer !
	}
	return (0);								// 0 als Fehlanzeige zurckliefern
}

//------------------------------------------------------------------------------
//  SearchTxt:  Direkte Mustersuche in einem Text
//  Parameter:  s   String, in dem gesucht wird
//              p   Suchmuster
//------------------------------------------------------------------------------
int  SearchTxt  (char *s, char *p)
{
	int n, m;								// String- Musterlnge
	int i, j;								// Suchzeiger
	unsigned  int found;					// Flag:  1 == gefunden
											//        0 == nicht gefunden
	n = strlen(s);
	m = strlen(p);
	for (i=0; i<n-m+1; i++)
	{
		found = 1;
		for (j=0; j<m; j++)
			if (s[i+j] != p[j]) found = 0;
		if (found == 1) return (i);
	}
	return(-1);
}

//------------------------------------------------------------------------------
//  SearchKMP_O:  Mustersuche nach den Knuth-Morris-Pratt-Algorithmus
//                (nach Ottmann)
//  Parameter:    s   String, in dem gesucht wird
//                p   Suchmuster
//------------------------------------------------------------------------------
int  SearchKMP_O  (char *s, char *p)
{
    int i, j;								// Suchzeiger
    int n = strlen(s);						// Stringlnge
    int m = strlen(p);						// Musterlnge
    int next[pMax];                         // Zeigerfeld fr die KMP-Suche

	InitNext_O(p,next);

    i=0; j=0;
    do
    {
       if ( j<=0 || (s[i]==p[j]) )
           { i++; j++; }
       else
          j=next[j];
    } while ( (j<m) && (i<n) );
    if (j>=m) return (i-m);
    else      return (-1);
}

void InitNext_O (char *p, int next[]) 	    // Vorbereitung: Initialisiere das Feld next
{
	int i, j;								// Laufzeiger
	int m = strlen(p);						// Musterlnge

    i=0; j=-1;
	next[i]=-1;
	do
    {
       if ( j<0 || (p[i]==p[j]) )
          { i++; j++; next[i]=j; }
       else
          j=next[j];
    } while (i<m-1);
    
}

//------------------------------------------------------------------------------
//  SearchKMP_S:  Mustersuche nach den Knuth-Morris-Pratt-Algorithmus
//                nach Sedgewick
//  Parameter:    s   String, in dem gesucht wird
//                p   Suchmuster
//------------------------------------------------------------------------------
int  SearchKMP_S  (char *s, char *p)
{
	int i, j;								// Suchzeiger
	int n = strlen(s);						// Stringlnge
	int m = strlen(p);						// Musterlnge
   int  next[pMax];                         // Zeigerfeld fr die KMP-Suche

	InitNext_S(p,next);
	for (i=0, j=0; j<m && i<n; i++, j++)
		while ((j>=0) && (s[i] != p[j]))  j=next[j];
	if (j==m) 	return (i-m);
	else  		return (-1);
}

void InitNext_S (char *p, int next[]) 	    // Vorbereitung: Initialisiere das Feld next
{
	int i, j;								// Laufzeiger
	int m = strlen(p);						// Musterlnge

	next[0] = -1;
	for (i=0, j=-1;i<m;i++, j++, next[i]=j)
		while ((j>=0) && (p[i] != p[j]))  j=next[j];
}

//------------------------------------------------------------------------------
//  Levenstein :  Berechnen der Levenstein-Distanz zweier Worte
//  Parameter     wort 1        erstes Vergleichswort
//                wort 2        zweites Vergleichswort
//                Return-Wert   Levenstein-Dustanz der beiden Worte
//------------------------------------------------------------------------------

// Minimum von 3 Zahlen:
inline int min3(int x,int y,int z)
       { if(x<y) y=x; if(y<z) z=y; return(z); }

// Kosten der Ersetzung eines Buchstabens an Position (x,y)
inline int pp(char* wort1, char* wort2, int x, int y)
       { if(wort1[x] == wort2[y]) return(0);  else return(p); }

int LvDist(char* wort1, char* wort2)
{
    int i, j, d[wMax+1][wMax+1];

    d[0][0] = 0;
    for(j=1;j<=wMax;j++)  d[0][j]=d[0][j-1] + q;
    for(i=1;i<=wMax;i++)  d[i][0]=d[i - 1][0] + r;
    for(i=1;i<=strlen(wort1);i++)
      for(j=1;j<=strlen(wort2);j++)
        d[i][j] = min3( d[i-1][j-1] + pp(wort1,wort2,i - 1,j - 1),
                        d[i][j-1] + q,
		                d[i-1][j] + r);

    // Kontrollausgabe der Levenshtein-Matrix
    // for(i=1;i<=strlen(wort1);i++)
    // {
    //   for(j=1;j<=strlen(wort2);j++)
    //       cout << d[i][j] << "   ";
    //   cout << endl;
    // }
    return(d[strlen(wort1)][strlen(wort2)]);
}

//------------------------------------------------------------------------------
//  SearchLv:   Direkte Mustersuche in einer Textzeile
//  Parameter:  line       String, in dem gesucht wird
//              pattern    Suchmuster
//              tolerance  Toleranz fr die Levebnstein-Distanz
//              pos        Position des gefundenen Musters in s, beginnend mit 1
//                         oder -1 bei Fehlanzeige
//------------------------------------------------------------------------------
void  SearchLv (char *line, char *pattern, int tolerance, int &pos)
{
    int  dist;
    char *word;

    if (strlen(line)==0) { pos = -1; return; }
    word = strtok (line, ",.:- \n");     // Trennzeichen sind:  ,.- \n
    do
    {
      dist = LvDist (word, pattern);
      if (dist <= tolerance) { pos = word - line; return; }
      word = strtok (NULL, ",.-:;()[]{} \n");
    } while (word!=NULL);
	pos = -1;
    return;
}


